@@ -19,7 +19,7 @@
 //! CrdsValue enums.
 //!
 //! Merge strategy is implemented in:
-//!     impl PartialOrd for VersionedCrdsValue
+//!     fn overrides(value: &CrdsValue, other: &VersionedCrdsValue) -> bool
 //!
 //! A value is updated to a new version if the labels match, and the value
 //! wallclock is later, or the value hash is greater.
@@ -35,14 +35,11 @@ use solana_sdk::hash::{hash, Hash};
 use solana_sdk::pubkey::Pubkey;
 use std::{
     cmp::Ordering,
-    collections::{hash_map, BTreeMap, HashMap},
+    collections::{hash_map, BTreeMap, HashMap, VecDeque},
     ops::{Bound, Index, IndexMut},
 };
 
 const CRDS_SHARDS_BITS: u32 = 8;
-// Limit number of crds values associated with each unique pubkey. This
-// excludes crds values which by label design are limited per each pubkey.
-const MAX_CRDS_VALUES_PER_PUBKEY: usize = 32;
 
 #[derive(Clone)]
 pub struct Crds {
@@ -51,26 +48,25 @@ pub struct Crds {
     cursor: Cursor, // Next insert ordinal location.
     shards: CrdsShards,
     nodes: IndexSet<usize>, // Indices of nodes' ContactInfo.
-    votes: IndexSet<usize>, // Indices of Vote crds values.
+    // Indices of Votes keyed by insert order.
+    votes: BTreeMap<u64 /*insert order*/, usize /*index*/>,
     // Indices of EpochSlots keyed by insert order.
     epoch_slots: BTreeMap<u64 /*insert order*/, usize /*index*/>,
     // Indices of all crds values associated with a node.
     records: HashMap<Pubkey, IndexSet<usize>>,
     // Indices of all entries keyed by insert order.
     entries: BTreeMap<u64 /*insert order*/, usize /*index*/>,
+    // Hash of recently purged values.
+    purged: VecDeque<(Hash, u64 /*timestamp*/)>,
 }
 
 #[derive(PartialEq, Debug)]
 pub enum CrdsError {
-    // Hash of the crds value which failed to insert should be recorded in
-    // failed_inserts to be excluded from the next pull-request.
-    InsertFailed(Hash),
+    InsertFailed,
     UnknownStakes,
 }
 
 /// This structure stores some local metadata associated with the CrdsValue
-/// The implementation of PartialOrd ensures that the "highest" version is always picked to be
-/// stored in the Crds
 #[derive(PartialEq, Debug, Clone)]
 pub struct VersionedCrdsValue {
     /// Ordinal index indicating insert order.
@@ -116,10 +112,11 @@ impl Default for Crds {
             cursor: Cursor::default(),
             shards: CrdsShards::new(CRDS_SHARDS_BITS),
             nodes: IndexSet::default(),
-            votes: IndexSet::default(),
+            votes: BTreeMap::default(),
             epoch_slots: BTreeMap::default(),
             records: HashMap::default(),
             entries: BTreeMap::default(),
+            purged: VecDeque::default(),
         }
     }
 }
@@ -128,6 +125,14 @@ impl Default for Crds {
 // Both values should have the same key/label.
 fn overrides(value: &CrdsValue, other: &VersionedCrdsValue) -> bool {
     assert_eq!(value.label(), other.value.label(), "labels mismatch!");
+    // Node instances are special cased so that if there are two running
+    // instances of the same node, the more recent start is propagated through
+    // gossip regardless of wallclocks.
+    if let CrdsData::NodeInstance(value) = &value.data {
+        if let Some(out) = value.overrides(&other.value) {
+            return out;
+        }
+    }
     match value.wallclock().cmp(&other.value.wallclock()) {
         Ordering::Less => false,
         Ordering::Greater => true,
@@ -152,14 +157,10 @@ impl Crds {
         }
     }
 
-    pub fn insert(
-        &mut self,
-        value: CrdsValue,
-        local_timestamp: u64,
-    ) -> Result<Option<VersionedCrdsValue>, CrdsError> {
+    pub fn insert(&mut self, value: CrdsValue, now: u64) -> Result<(), CrdsError> {
         let label = value.label();
         let pubkey = value.pubkey();
-        let value = VersionedCrdsValue::new(value, self.cursor, local_timestamp);
+        let value = VersionedCrdsValue::new(value, self.cursor, now);
         match self.table.entry(label) {
             Entry::Vacant(entry) => {
                 let entry_index = entry.index();
@@ -169,7 +170,7 @@ impl Crds {
                         self.nodes.insert(entry_index);
                     }
                     CrdsData::Vote(_, _) => {
-                        self.votes.insert(entry_index);
+                        self.votes.insert(value.ordinal, entry_index);
                     }
                     CrdsData::EpochSlots(_, _) => {
                         self.epoch_slots.insert(value.ordinal, entry_index);
@@ -180,15 +181,22 @@ impl Crds {
                 self.records.entry(pubkey).or_default().insert(entry_index);
                 self.cursor.consume(value.ordinal);
                 entry.insert(value);
-                Ok(None)
+                Ok(())
             }
             Entry::Occupied(mut entry) if overrides(&value.value, entry.get()) => {
                 let entry_index = entry.index();
                 self.shards.remove(entry_index, entry.get());
                 self.shards.insert(entry_index, &value);
-                if let CrdsData::EpochSlots(_, _) = value.value.data {
-                    self.epoch_slots.remove(&entry.get().ordinal);
-                    self.epoch_slots.insert(value.ordinal, entry_index);
+                match value.value.data {
+                    CrdsData::Vote(_, _) => {
+                        self.votes.remove(&entry.get().ordinal);
+                        self.votes.insert(value.ordinal, entry_index);
+                    }
+                    CrdsData::EpochSlots(_, _) => {
+                        self.epoch_slots.remove(&entry.get().ordinal);
+                        self.epoch_slots.insert(value.ordinal, entry_index);
+                    }
+                    _ => (),
                 }
                 self.entries.remove(&entry.get().ordinal);
                 self.entries.insert(value.ordinal, entry_index);
@@ -196,27 +204,24 @@ impl Crds {
                 // does not need to be updated.
                 debug_assert_eq!(entry.get().value.pubkey(), pubkey);
                 self.cursor.consume(value.ordinal);
-                Ok(Some(entry.insert(value)))
+                self.purged.push_back((entry.get().value_hash, now));
+                entry.insert(value);
+                Ok(())
             }
-            _ => {
+            Entry::Occupied(entry) => {
                 trace!(
                     "INSERT FAILED data: {} new.wallclock: {}",
                     value.value.label(),
                     value.value.wallclock(),
                 );
-                Err(CrdsError::InsertFailed(value.value_hash))
+                if entry.get().value_hash != value.value_hash {
+                    self.purged.push_back((value.value_hash, now));
+                }
+                Err(CrdsError::InsertFailed)
             }
         }
     }
 
-    pub fn lookup(&self, label: &CrdsValueLabel) -> Option<&CrdsValue> {
-        self.table.get(label).map(|x| &x.value)
-    }
-
-    pub fn lookup_versioned(&self, label: &CrdsValueLabel) -> Option<&VersionedCrdsValue> {
-        self.table.get(label)
-    }
-
     pub fn get(&self, label: &CrdsValueLabel) -> Option<&VersionedCrdsValue> {
         self.table.get(label)
     }
@@ -250,15 +255,10 @@ impl Crds {
         &'a self,
         cursor: &'a mut Cursor,
     ) -> impl Iterator<Item = &'a VersionedCrdsValue> {
-        let since = cursor.ordinal();
-        self.votes.iter().filter_map(move |i| {
-            let entry = self.table.index(*i);
-            if entry.ordinal >= since {
-                cursor.consume(entry.ordinal);
-                Some(entry)
-            } else {
-                None
-            }
+        let range = (Bound::Included(cursor.ordinal()), Bound::Unbounded);
+        self.votes.range(range).map(move |(ordinal, index)| {
+            cursor.consume(*ordinal);
+            self.table.index(*index)
         })
     }
 
@@ -318,6 +318,24 @@ impl Crds {
         self.table.par_values()
     }
 
+    pub(crate) fn num_purged(&self) -> usize {
+        self.purged.len()
+    }
+
+    pub(crate) fn purged(&self) -> impl IndexedParallelIterator<Item = Hash> + '_ {
+        self.purged.par_iter().map(|(hash, _)| *hash)
+    }
+
+    /// Drops purged value hashes with timestamp less than the given one.
+    pub(crate) fn trim_purged(&mut self, timestamp: u64) {
+        let count = self
+            .purged
+            .iter()
+            .take_while(|(_, ts)| *ts < timestamp)
+            .count();
+        self.purged.drain(..count);
+    }
+
     /// Returns all crds values which the first 'mask_bits'
     /// of their hash value is equal to 'mask'.
     pub fn filter_bitmask(
@@ -372,9 +390,7 @@ impl Crds {
                     None => 0,
                 }
             };
-            let mut old_labels = Vec::new();
-            // Buffer of crds values to be evicted based on their wallclock.
-            let mut recent_unlimited_labels: Vec<(u64 /*wallclock*/, usize /*index*/)> = index
+            index
                 .into_iter()
                 .filter_map(|ix| {
                     let (label, value) = self.table.get_index(*ix).unwrap();
@@ -383,32 +399,12 @@ impl Crds {
                         .max(local_timestamp)
                         .saturating_add(timeout);
                     if expiry_timestamp <= now {
-                        old_labels.push(label.clone());
-                        None
+                        Some(label.clone())
                     } else {
-                        match label.value_space() {
-                            Some(_) => None,
-                            None => Some((value.value.wallclock(), *ix)),
-                        }
+                        None
                     }
                 })
-                .collect();
-            // Number of values to discard from the buffer:
-            let nth = recent_unlimited_labels
-                .len()
-                .saturating_sub(MAX_CRDS_VALUES_PER_PUBKEY);
-            // Partition on wallclock to discard the older ones.
-            if nth > 0 && nth < recent_unlimited_labels.len() {
-                recent_unlimited_labels.select_nth_unstable(nth);
-            }
-            old_labels.extend(
-                recent_unlimited_labels
-                    .split_at(nth)
-                    .0
-                    .iter()
-                    .map(|(_ /*wallclock*/, ix)| self.table.get_index(*ix).unwrap().0.clone()),
-            );
-            old_labels
+                .collect::<Vec<_>>()
         };
         thread_pool.install(|| {
             self.records
@@ -418,15 +414,19 @@ impl Crds {
         })
     }
 
-    pub fn remove(&mut self, key: &CrdsValueLabel) -> Option<VersionedCrdsValue> {
-        let (index, _ /*label*/, value) = self.table.swap_remove_full(key)?;
+    pub fn remove(&mut self, key: &CrdsValueLabel, now: u64) {
+        let (index, _ /*label*/, value) = match self.table.swap_remove_full(key) {
+            Some(entry) => entry,
+            None => return,
+        };
+        self.purged.push_back((value.value_hash, now));
         self.shards.remove(index, &value);
         match value.value.data {
             CrdsData::ContactInfo(_) => {
                 self.nodes.swap_remove(&index);
             }
             CrdsData::Vote(_, _) => {
-                self.votes.swap_remove(&index);
+                self.votes.remove(&value.ordinal);
             }
             CrdsData::EpochSlots(_, _) => {
                 self.epoch_slots.remove(&value.ordinal);
@@ -460,8 +460,7 @@ impl Crds {
                     self.nodes.insert(index);
                 }
                 CrdsData::Vote(_, _) => {
-                    self.votes.swap_remove(&size);
-                    self.votes.insert(index);
+                    self.votes.insert(value.ordinal, index);
                 }
                 CrdsData::EpochSlots(_, _) => {
                     self.epoch_slots.insert(value.ordinal, index);
@@ -474,7 +473,6 @@ impl Crds {
             records.swap_remove(&size);
             records.insert(index);
         }
-        Some(value)
     }
 
     /// Returns true if the number of unique pubkeys in the table exceeds the
@@ -495,12 +493,13 @@ impl Crds {
         // e.g. trusted validators, self pubkey, ...
         keep: &[Pubkey],
         stakes: &HashMap<Pubkey, u64>,
-    ) -> Result<Vec<VersionedCrdsValue>, CrdsError> {
+        now: u64,
+    ) -> Result</*num purged:*/ usize, CrdsError> {
         if self.should_trim(cap) {
             let size = self.records.len().saturating_sub(cap);
-            self.drop(size, keep, stakes)
+            self.drop(size, keep, stakes, now)
         } else {
-            Ok(Vec::default())
+            Ok(0)
         }
     }
 
@@ -510,7 +509,8 @@ impl Crds {
         size: usize,
         keep: &[Pubkey],
         stakes: &HashMap<Pubkey, u64>,
-    ) -> Result<Vec<VersionedCrdsValue>, CrdsError> {
+        now: u64,
+    ) -> Result</*num purged:*/ usize, CrdsError> {
         if stakes.is_empty() {
             return Err(CrdsError::UnknownStakes);
         }
@@ -530,7 +530,10 @@ impl Crds {
             .flat_map(|k| &self.records[&k])
             .map(|k| self.table.get_index(*k).unwrap().0.clone())
             .collect();
-        Ok(keys.iter().map(|k| self.remove(k).unwrap()).collect())
+        for key in &keys {
+            self.remove(key, now);
+        }
+        Ok(keys.len())
     }
 }
 
@@ -541,7 +544,8 @@ mod test {
         contact_info::ContactInfo,
         crds_value::{new_rand_timestamp, NodeInstance},
     };
-    use rand::{thread_rng, Rng};
+    use rand::{thread_rng, Rng, SeedableRng};
+    use rand_chacha::ChaChaRng;
     use rayon::ThreadPoolBuilder;
     use solana_sdk::signature::{Keypair, Signer};
     use std::{collections::HashSet, iter::repeat_with};
@@ -550,7 +554,7 @@ mod test {
     fn test_insert() {
         let mut crds = Crds::default();
         let val = CrdsValue::new_unsigned(CrdsData::ContactInfo(ContactInfo::default()));
-        assert_eq!(crds.insert(val.clone(), 0).ok(), Some(None));
+        assert_eq!(crds.insert(val.clone(), 0), Ok(()));
         assert_eq!(crds.table.len(), 1);
         assert!(crds.table.contains_key(&val.label()));
         assert_eq!(crds.table[&val.label()].local_timestamp, 0);
@@ -559,12 +563,9 @@ mod test {
     fn test_update_old() {
         let mut crds = Crds::default();
         let val = CrdsValue::new_unsigned(CrdsData::ContactInfo(ContactInfo::default()));
-        let value_hash = hash(&serialize(&val).unwrap());
-        assert_eq!(crds.insert(val.clone(), 0), Ok(None));
-        assert_eq!(
-            crds.insert(val.clone(), 1),
-            Err(CrdsError::InsertFailed(value_hash))
-        );
+        assert_eq!(crds.insert(val.clone(), 0), Ok(()));
+        assert_eq!(crds.insert(val.clone(), 1), Err(CrdsError::InsertFailed));
+        assert!(crds.purged.is_empty());
         assert_eq!(crds.table[&val.label()].local_timestamp, 0);
     }
     #[test]
@@ -574,15 +575,14 @@ mod test {
             &Pubkey::default(),
             0,
         )));
-        assert_matches!(crds.insert(original.clone(), 0), Ok(_));
+        let value_hash = hash(&serialize(&original).unwrap());
+        assert_matches!(crds.insert(original, 0), Ok(()));
         let val = CrdsValue::new_unsigned(CrdsData::ContactInfo(ContactInfo::new_localhost(
             &Pubkey::default(),
             1,
         )));
-        assert_eq!(
-            crds.insert(val.clone(), 1).unwrap().unwrap().value,
-            original
-        );
+        assert_eq!(crds.insert(val.clone(), 1), Ok(()));
+        assert_eq!(*crds.purged.back().unwrap(), (value_hash, 1));
         assert_eq!(crds.table[&val.label()].local_timestamp, 1);
     }
     #[test]
@@ -592,12 +592,13 @@ mod test {
             &Pubkey::default(),
             0,
         )));
-        assert_eq!(crds.insert(val.clone(), 0), Ok(None));
+        assert_eq!(crds.insert(val.clone(), 0), Ok(()));
         assert_eq!(crds.table[&val.label()].ordinal, 0);
 
         let val2 = CrdsValue::new_unsigned(CrdsData::ContactInfo(ContactInfo::default()));
+        let value_hash = hash(&serialize(&val2).unwrap());
         assert_eq!(val2.label().pubkey(), val.label().pubkey());
-        assert_matches!(crds.insert(val2.clone(), 0), Ok(Some(_)));
+        assert_eq!(crds.insert(val2.clone(), 0), Ok(()));
 
         crds.update_record_timestamp(&val.label().pubkey(), 2);
         assert_eq!(crds.table[&val.label()].local_timestamp, 2);
@@ -612,16 +613,73 @@ mod test {
         let mut ci = ContactInfo::default();
         ci.wallclock += 1;
         let val3 = CrdsValue::new_unsigned(CrdsData::ContactInfo(ci));
-        assert_matches!(crds.insert(val3, 3), Ok(Some(_)));
+        assert_eq!(crds.insert(val3, 3), Ok(()));
+        assert_eq!(*crds.purged.back().unwrap(), (value_hash, 3));
         assert_eq!(crds.table[&val2.label()].local_timestamp, 3);
         assert_eq!(crds.table[&val2.label()].ordinal, 2);
     }
+
+    #[test]
+    fn test_upsert_node_instance() {
+        const SEED: [u8; 32] = [0x42; 32];
+        let mut rng = ChaChaRng::from_seed(SEED);
+        fn make_crds_value(node: NodeInstance) -> CrdsValue {
+            CrdsValue::new_unsigned(CrdsData::NodeInstance(node))
+        }
+        let now = 1_620_838_767_000;
+        let mut crds = Crds::default();
+        let pubkey = Pubkey::new_unique();
+        let node = NodeInstance::new(&mut rng, pubkey, now);
+        let node = make_crds_value(node);
+        assert_eq!(crds.insert(node, now), Ok(()));
+        // A node-instance with a different key should insert fine even with
+        // older timestamps.
+        let other = NodeInstance::new(&mut rng, Pubkey::new_unique(), now - 1);
+        let other = make_crds_value(other);
+        assert_eq!(crds.insert(other, now), Ok(()));
+        // A node-instance with older timestamp should fail to insert, even if
+        // the wallclock is more recent.
+        let other = NodeInstance::new(&mut rng, pubkey, now - 1);
+        let other = other.with_wallclock(now + 1);
+        let other = make_crds_value(other);
+        let value_hash = hash(&serialize(&other).unwrap());
+        assert_eq!(crds.insert(other, now), Err(CrdsError::InsertFailed));
+        assert_eq!(*crds.purged.back().unwrap(), (value_hash, now));
+        // A node instance with the same timestamp should insert only if the
+        // random token is larger.
+        let mut num_overrides = 0;
+        for _ in 0..100 {
+            let other = NodeInstance::new(&mut rng, pubkey, now);
+            let other = make_crds_value(other);
+            let value_hash = hash(&serialize(&other).unwrap());
+            match crds.insert(other, now) {
+                Ok(()) => num_overrides += 1,
+                Err(CrdsError::InsertFailed) => {
+                    assert_eq!(*crds.purged.back().unwrap(), (value_hash, now))
+                }
+                _ => panic!(),
+            }
+        }
+        assert_eq!(num_overrides, 5);
+        // A node instance with larger timestamp should insert regardless of
+        // its token value.
+        for k in 1..10 {
+            let other = NodeInstance::new(&mut rng, pubkey, now + k);
+            let other = other.with_wallclock(now - 1);
+            let other = make_crds_value(other);
+            match crds.insert(other, now) {
+                Ok(()) => (),
+                _ => panic!(),
+            }
+        }
+    }
+
     #[test]
     fn test_find_old_records_default() {
         let thread_pool = ThreadPoolBuilder::new().build().unwrap();
         let mut crds = Crds::default();
         let val = CrdsValue::new_unsigned(CrdsData::ContactInfo(ContactInfo::default()));
-        assert_eq!(crds.insert(val.clone(), 1), Ok(None));
+        assert_eq!(crds.insert(val.clone(), 1), Ok(()));
         let mut set = HashMap::new();
         set.insert(Pubkey::default(), 0);
         assert!(crds.find_old_labels(&thread_pool, 0, &set).is_empty());
@@ -644,7 +702,7 @@ mod test {
         let mut timeouts = HashMap::new();
         let val = CrdsValue::new_rand(&mut rng, None);
         timeouts.insert(Pubkey::default(), 3);
-        assert_eq!(crds.insert(val.clone(), 0), Ok(None));
+        assert_eq!(crds.insert(val.clone(), 0), Ok(()));
         assert!(crds.find_old_labels(&thread_pool, 2, &timeouts).is_empty());
         timeouts.insert(val.pubkey(), 1);
         assert_eq!(
@@ -662,40 +720,6 @@ mod test {
         );
     }
 
-    #[test]
-    fn test_find_old_records_unlimited() {
-        let thread_pool = ThreadPoolBuilder::new().build().unwrap();
-        let mut rng = thread_rng();
-        let now = 1_610_034_423_000;
-        let pubkey = Pubkey::new_unique();
-        let mut crds = Crds::default();
-        let mut timeouts = HashMap::new();
-        timeouts.insert(Pubkey::default(), 1);
-        timeouts.insert(pubkey, 180);
-        for _ in 0..1024 {
-            let wallclock = now - rng.gen_range(0, 240);
-            let val = NodeInstance::new(&mut rng, pubkey, wallclock);
-            let val = CrdsData::NodeInstance(val);
-            let val = CrdsValue::new_unsigned(val);
-            assert_eq!(crds.insert(val, now), Ok(None));
-        }
-        let now = now + 1;
-        let labels = crds.find_old_labels(&thread_pool, now, &timeouts);
-        assert_eq!(crds.table.len() - labels.len(), MAX_CRDS_VALUES_PER_PUBKEY);
-        let max_wallclock = labels
-            .iter()
-            .map(|label| crds.lookup(label).unwrap().wallclock())
-            .max()
-            .unwrap();
-        assert!(max_wallclock > now - 180);
-        let labels: HashSet<_> = labels.into_iter().collect();
-        for (label, value) in crds.table.iter() {
-            if !labels.contains(label) {
-                assert!(max_wallclock <= value.value.wallclock());
-            }
-        }
-    }
-
     #[test]
     fn test_remove_default() {
         let thread_pool = ThreadPoolBuilder::new().build().unwrap();
@@ -708,7 +732,7 @@ mod test {
             crds.find_old_labels(&thread_pool, 2, &set),
             vec![val.label()]
         );
-        crds.remove(&val.label());
+        crds.remove(&val.label(), /*now=*/ 0);
         assert!(crds.find_old_labels(&thread_pool, 2, &set).is_empty());
     }
     #[test]
@@ -716,7 +740,7 @@ mod test {
         let thread_pool = ThreadPoolBuilder::new().build().unwrap();
         let mut crds = Crds::default();
         let val = CrdsValue::new_unsigned(CrdsData::ContactInfo(ContactInfo::default()));
-        assert_eq!(crds.insert(val.clone(), 1), Ok(None));
+        assert_eq!(crds.insert(val.clone(), 1), Ok(()));
         let mut set = HashMap::new();
         //now < timestamp
         set.insert(Pubkey::default(), 0);
@@ -754,27 +778,19 @@ mod test {
         let keypairs: Vec<_> = std::iter::repeat_with(Keypair::new).take(256).collect();
         let mut rng = thread_rng();
         let mut num_inserts = 0;
-        let mut num_overrides = 0;
         for _ in 0..4096 {
             let keypair = &keypairs[rng.gen_range(0, keypairs.len())];
             let value = CrdsValue::new_rand(&mut rng, Some(keypair));
             let local_timestamp = new_rand_timestamp(&mut rng);
-            match crds.insert(value, local_timestamp) {
-                Ok(None) => {
-                    num_inserts += 1;
-                    check_crds_shards(&crds);
-                }
-                Ok(Some(_)) => {
-                    num_inserts += 1;
-                    num_overrides += 1;
-                    check_crds_shards(&crds);
-                }
-                Err(_) => (),
+            if let Ok(()) = crds.insert(value, local_timestamp) {
+                num_inserts += 1;
+                check_crds_shards(&crds);
             }
         }
         assert_eq!(num_inserts, crds.cursor.0 as usize);
         assert!(num_inserts > 700);
-        assert!(num_overrides > 500);
+        assert!(crds.num_purged() > 500);
+        assert_eq!(crds.num_purged() + crds.table.len(), 4096);
         assert!(crds.table.len() > 200);
         assert!(num_inserts > crds.table.len());
         check_crds_shards(&crds);
@@ -782,7 +798,7 @@ mod test {
         while !crds.table.is_empty() {
             let index = rng.gen_range(0, crds.table.len());
             let key = crds.table.get_index(index).unwrap().0.clone();
-            crds.remove(&key);
+            crds.remove(&key, /*now=*/ 0);
             check_crds_shards(&crds);
         }
     }
@@ -916,20 +932,12 @@ mod test {
         let keypairs: Vec<_> = repeat_with(Keypair::new).take(128).collect();
         let mut crds = Crds::default();
         let mut num_inserts = 0;
-        let mut num_overrides = 0;
         for k in 0..4096 {
             let keypair = &keypairs[rng.gen_range(0, keypairs.len())];
             let value = CrdsValue::new_rand(&mut rng, Some(keypair));
             let local_timestamp = new_rand_timestamp(&mut rng);
-            match crds.insert(value, local_timestamp) {
-                Ok(None) => {
-                    num_inserts += 1;
-                }
-                Ok(Some(_)) => {
-                    num_inserts += 1;
-                    num_overrides += 1;
-                }
-                Err(_) => (),
+            if let Ok(()) = crds.insert(value, local_timestamp) {
+                num_inserts += 1;
             }
             if k % 16 == 0 {
                 check_crds_value_indices(&mut rng, &crds);
@@ -937,8 +945,9 @@ mod test {
         }
         assert_eq!(num_inserts, crds.cursor.0 as usize);
         assert!(num_inserts > 700);
-        assert!(num_overrides > 500);
+        assert!(crds.num_purged() > 500);
         assert!(crds.table.len() > 200);
+        assert_eq!(crds.num_purged() + crds.table.len(), 4096);
         assert!(num_inserts > crds.table.len());
         let (num_nodes, num_votes, num_epoch_slots) = check_crds_value_indices(&mut rng, &crds);
         assert!(num_nodes * 3 < crds.table.len());
@@ -953,7 +962,7 @@ mod test {
         while !crds.table.is_empty() {
             let index = rng.gen_range(0, crds.table.len());
             let key = crds.table.get_index(index).unwrap().0.clone();
-            crds.remove(&key);
+            crds.remove(&key, /*now=*/ 0);
             if crds.table.len() % 16 == 0 {
                 check_crds_value_indices(&mut rng, &crds);
             }
@@ -992,7 +1001,7 @@ mod test {
         while !crds.table.is_empty() {
             let index = rng.gen_range(0, crds.table.len());
             let key = crds.table.get_index(index).unwrap().0.clone();
-            crds.remove(&key);
+            crds.remove(&key, /*now=*/ 0);
             if crds.table.len() % 64 == 0 {
                 check_crds_records(&crds);
             }
@@ -1001,6 +1010,7 @@ mod test {
     }
 
     #[test]
+    #[allow(clippy::needless_collect)]
     fn test_drop() {
         fn num_unique_pubkeys<'a, I>(values: I) -> usize
         where
@@ -1029,7 +1039,15 @@ mod test {
         let num_pubkeys = num_unique_pubkeys(crds.table.values());
         assert!(!crds.should_trim(num_pubkeys));
         assert!(crds.should_trim(num_pubkeys * 5 / 6));
-        let purged = crds.drop(16, &[], &stakes).unwrap();
+        let values: Vec<_> = crds.table.values().cloned().collect();
+        crds.drop(16, &[], &stakes, /*now=*/ 0).unwrap();
+        let purged: Vec<_> = {
+            let purged: HashSet<_> = crds.purged.iter().map(|(hash, _)| hash).copied().collect();
+            values
+                .into_iter()
+                .filter(|v| purged.contains(&v.value_hash))
+                .collect()
+        };
         assert_eq!(purged.len() + crds.table.len(), num_values);
         assert_eq!(num_unique_pubkeys(&purged), 16);
         assert_eq!(num_unique_pubkeys(crds.table.values()), num_pubkeys - 16);
@@ -1066,7 +1084,7 @@ mod test {
             crds.find_old_labels(&thread_pool, 2, &set),
             vec![val.label()]
         );
-        crds.remove(&val.label());
+        crds.remove(&val.label(), /*now=*/ 0);
         assert!(crds.find_old_labels(&thread_pool, 2, &set).is_empty());
     }
 
